
#include "cnf.h"
#include <stdexcept>
#include <unordered_set>


geometric::constraints::cnf::cnf( 
   std::initializer_list< std::initializer_list< int >> init )
{
   for( auto cl : init )  
   {
      clauseset. push_back( clause( ) );
   
      for( auto lit : cl )
         clauseset. back( ). push_back( lit ); 

   }
}


int geometric::constraints::cnf::maxvar( ) const
{
   int res = 0;
   for( const auto& cl : clauseset )
      for( const auto lit : cl )
      {
         if( lit == 0 )
            throw std::runtime_error( "variable 0 not allowed" );

         if( lit > 0 && lit > res )
            res = lit;
         else
         {
            if( lit < 0 && -lit > res )
               res = -lit;
         }
      }

   return res;
}


std::ostream& 
geometric::constraints::operator << ( std::ostream& out, const cnf& c )
{
   out << "c  cnf generated by Geo\n";
   out << "c  Written by Hans de Nivelle, 2017\n";
   out << "\n";

   out << "p cnf " << c. maxvar( ) << " " << c. nrclauses( ) << "\n";

   for( const auto& cl : c. clauseset )
   {
      for( const auto lit : cl )
         out << lit << " ";
      out << "0\n";
   }

   return out;
}


namespace
{

   // Make an atom from g. clauses[cl] / i. 
   // (We lookup in firstliterals where cl starts, and add i to it.) 

   int make_atom( const std::vector< size_t > & firstliterals, 
                  const geometric::constraints::gcsp& g, 
                  size_t cl, size_t i )
   {
      if( cl >= g. clauses. size( ))
         throw std::logic_error( "getnumber: clause cl too high" );

      if( !g. clauses[cl]. current. contains(i) )
         throw std::logic_error( "getnumber: choice not in current" );

      return firstliterals[ cl ] + ( i - g. clauses[cl]. current. s0 );
   }
}


geometric::constraints::cnf 
geometric::constraints::translate2cnf( const gcsp& g )
{
   cnf result;

   std::vector< size_t > firstliterals; 
      // For the cl-th clause, firstliterals[cl] is the integer 
      // that represents the atom that represents the first substlet. 

   size_t currentliteral = 1; 
   for( size_t cl = 0; cl < g. clauses. size( ); ++ cl )
   {
      cnf::clause choices;   // Translation of g. clauses[cl];

      firstliterals. push_back( currentliteral );

      for( size_t i = g. clauses[cl]. current. s0;
                  i != g. clauses[cl]. current. s1;
                  ++ i )
      {
         choices. push_back( currentliteral ++ ); 
      } 

      result. insert( std::move( choices ));
   }
   
   firstliterals. push_back( currentliteral );
      // Now we have one more.

   // We look for the clashing substlets within the clauses,
   // and add the corresponding literals: 

   for( size_t cl1 = 0; cl1 < g. clauses. size( ); ++ cl1 )
   {
      std::unordered_set< size_t > checked;
         // Clauses that we checked against. Repetitions would be
         // bad for efficiency.

      for( size_t v = 0; v < g. clauses[cl1]. nrvars( ); ++ v )
      {
         const auto& occ = g. findclauses( g. clauses[cl1]. var(v) );
         for( const auto cl2 : occ )
         {
            // Conflicts are symmetric, and we don't care
            // about self conflicts.

            if( cl1 < cl2 && checked. count( cl2 ) == 0 ) 
               checked. insert( cl2 ); 
         }
      }

      for( auto cl2 : checked )
      {
         for( size_t i1 = g. clauses[ cl1 ]. current. s0; 
                     i1 < g. clauses[ cl1 ]. current. s1; ++ i1 )
         {
            for( size_t i2 = g. clauses[ cl2 ]. current. s0;
                        i2 < g. clauses[ cl2 ]. current. s1; ++ i2 )
            {
               // Check if there is a conflict, if yes, generate a clause
               // of form { -s1, -s2 }.

               if( !consistent( g. clauses[cl1], i1, g. clauses[cl2], i2 ))
               {
                  cnf::clause conflict =
                     { -make_atom( firstliterals, g, cl1, i1 ),
                       -make_atom( firstliterals, g, cl2, i2 ) };
               
                  result. insert( std::move( conflict ));  
               }
            }
         }
      }
   }

   for( size_t bl = 0; bl < g. blockings. size( ); ++ bl )
   {
      for( size_t j =  g. blockings[ bl ]. current. s0;
                  j != g. blockings[ bl ]. current. s1;
                  ++ j ) 
                 
      {
         // We need a set of clauses (from the gcsp) that cover the
         // variables in the blocking. The (propositional) clause that
         // we generate, consists of the substlets that are not consistent
         // with the blocking. 

         std::unordered_set< variable, variable::hash, variable::equal_to > 
                  covered;
        
         cnf::clause block;
 
         for( size_t v = 0; v < g. blockings[ bl ]. nrvars( ); ++ v )
         {
            variable var = g. blockings[ bl ]. var(v);
            size_t val = g. blockings[ bl ]. value( v, j );

            if( covered. count( var ) == 0 )
            {
               // best will be the shortest clause containing var:
               // One could consider using other heuristics.

               const auto& occ = g. findclauses( var );
               size_t best = g. clauses. size( );
 
               for( const auto i : occ )
               {
                  if( best == g. clauses. size( ) ||
                         g. clauses[ i ]. current. size( ) < 
                         g. clauses[ best ]. current. size( ))
                  {
                     best = i;
                  } 
               }

               size_t skipped = 0;
               for( size_t i = g. clauses[ best ]. current. s0;
                           i != g. clauses[ best ]. current. s1;
                           ++ i )
               {
                  // choices in i that are consistent with the blocking bl/j
                  // will not go into block. 

                  if( consistent( g. clauses[ best ], i,
                                  g. blockings[ bl ], j ))
                  {
                     ++ skipped;  
                  }
                  else
                     block. push_back( make_atom( firstliterals, g, best, i )); 
               } 

               // If there are no skipped (not entered) substlets, 
               // then it is not possible that the blocking will be 
               // ever implied, and there is no point adding a 
               // clause for it. So we jump away. 

               if( !skipped )
                  goto blockingcannothappen; 

               // Variables in clauses[ best ] are now covered.

               for( size_t v = 0; v < g. clauses[ best ]. nrvars( ); ++ v )
                  covered. insert( g. clauses[ best ]. var(v) );
            }
         }

         result. insert( std::move( block )); 
      blockingcannothappen:
         ;
      }
   }
   return result; 
}

